#! python3

"""
Common metric interface.
"""
from __future__ import annotations
from abc import ABC, abstractmethod
from dataclasses import dataclass
import importlib
import logging
import os
from typing import List, Optional

from dash import dcc
import dash_bootstrap_components as dbc
from dateutil.parser import isoparse
import pandas as pd
import plotly.graph_objects as go

#: Create logger for this file.
logger = logging.getLogger()


@dataclass
class MetricData:
    """
    This class is used to store all the tickets information used to produce
    metrics and analysis.
    """

    #: Raw data in case of advanced or custom processing needed
    raw: tuple

    #: Basics information like title, reporter, type, nb sprint, etc
    info: pd.DataFrame

    #: Information which change during time like status, assignee, etc
    changes: pd.DataFrame

    #: Workflow of the project
    workflow: list

    @classmethod
    def from_tickets_and_changelogs(
        cls,
        fields: dict,
        workflow: list,
        tickets: list,
        tickets_changelogs: list,
        tickets_parents: list,
    ) -> MetricData:
        """
        Create data for metric from `tickets` and `changelogs` information.

        :param fields: Dictionary of fields to retrieve.
        :param workflow: List of state in the workflow.
        :param tickets: List of tickets information.
        :param tickets_changelogs: List of changelogs associated to tickets.
        :param tickets_parents: List of parent tickets associated to tickets.
        :return: Data for metric.
        """
        logger.debug("Create data for metric from tickets and changelogs")

        raw = (fields, workflow, tickets, tickets_changelogs, tickets_parents)

        tickets_info = []
        tickets_changes = []
        status_backlog = cls._map_status_from_workflow("Backlog", workflow)

        for ticket, changelogs in zip(tickets, tickets_changelogs):
            key = ticket["key"]
            sprints = ticket["fields"].get(fields["sprint"])
            if not sprints:
                sprints = []

            fix_versions = []
            for fix_version in ticket["fields"]["fixVersions"]:
                fix_versions.append(fix_version["name"])

            parent = None
            if ticket["fields"].get("parent"):
                parent = ticket["fields"]["parent"]["key"]

            tickets_info.append(
                {
                    "key": key,
                    "type": ticket["fields"]["issuetype"]["name"],
                    "summary": ticket["fields"]["summary"],
                    "reporter": ticket["fields"]["reporter"]["displayName"],
                    "story point": ticket["fields"].get(
                        fields["story_points"]
                    ),
                    "sprints": len(sprints),
                    "versions": fix_versions,
                    "parent": parent,
                }
            )
            assignee = "Unassigned"
            status = status_backlog
            tickets_changes.append(
                {
                    "key": key,
                    "date": isoparse(ticket["fields"]["created"]),
                    "assignee": assignee,
                    "status": status,
                }
            )

            for changelog in changelogs:
                for change in changelog["items"]:
                    change_type = change.get("fieldId")
                    if not change_type:
                        continue
                    if change_type == "assignee":
                        if change["toString"]:
                            assignee = change["toString"]
                        else:
                            assignee = "Unassigned"
                    elif change_type == "status":
                        status = cls._map_status_from_workflow(
                            change["toString"], workflow
                        )
                    else:
                        continue

                    tickets_changes.append(
                        {
                            "key": key,
                            "date": isoparse(changelog["created"]),
                            "assignee": assignee,
                            "status": status,
                        }
                    )

        tickets_info_df = pd.DataFrame.from_records(tickets_info, index="key")
        tickets_changes_df = pd.DataFrame.from_records(tickets_changes)
        tickets_changes_df["date"] = pd.to_datetime(
            tickets_changes_df["date"], utc=True
        )
        status_order = pd.CategoricalDtype(
            categories=[state["name"] for state in workflow], ordered=True
        )
        tickets_changes_df["status"] = tickets_changes_df["status"].astype(
            status_order
        )

        logger.debug("Data for metric created")
        return cls(raw, tickets_info_df, tickets_changes_df, workflow)

    @staticmethod
    def _map_status_from_workflow(status: str, workflow: list) -> str:
        """
        Search the `status` associated in the `workflow` and map to the state
        name. If the `status` is not associated to a state, it returns the
        given `status` in input.

        :param status: Status to map.
        :param workflow: List of state in the workflow.
        :return: Status name.
        """
        for state in workflow:
            if status in state["status"]:
                return state["name"]
        return status


@dataclass
class MetricReport:
    """
    This class is used to store all the information used to produce report
    of the metric.
    """

    #: Name of the metric. Used to retrieve template associated.
    metric_name: str

    #: List of the figures to include in the report.
    figures: List[str]

    #: Additional data used to generate report.
    report_data: Optional[dict] = None


class Metric(ABC):
    """
    This class is used to provide an interface for all metrics.
    """

    #: Define output directory.
    OUTPUT_DIR: str = "out"

    @abstractmethod
    def compute_dashboard(self, data: MetricData) -> List[dbc.Row]:
        """
        Interface to compute widget of the dashboard related to this metric.

        :param data: Metric input data.
        :return: Dash widgets.
        """

    @abstractmethod
    def compute_report(self, data: MetricData) -> MetricReport:
        """
        Interface to compute widget of the report related to this metric.

        :param data: Metric input data.
        :return: Report of the metric.
        """

    @staticmethod
    def _draw_figure(title: str, figure: go.Figure) -> dbc.Col:
        """
        Helper function to set a graph in a common layout.

        :param title: Title of the graph.
        :param figure: Graph.
        :return: Dash widgets.
        """
        figure.update_traces(
            hoverinfo="label+percent", textinfo="value+percent"
        )
        return dbc.Col(
            dbc.Card(
                [
                    dbc.CardHeader(title),
                    dbc.CardBody(
                        dcc.Graph(
                            figure=figure,
                            config={
                                "displaylogo": False,
                            },
                        )
                    ),
                ]
            ),
            md=12,
            lg=6,
        )


class Metrics:
    """
    This class is used to manipulate all metrics registered.
    """

    def __init__(self, metrics: list):
        """
        Constructs the list of metric from the `metrics` list.

        :param metrics: List of metrics name.
        """
        self._metrics = Metrics._load_metrics(metrics)

        # Create output directory
        if not os.path.exists(Metric.OUTPUT_DIR):
            os.mkdir(Metric.OUTPUT_DIR)

    @staticmethod
    def _load_metrics(metrics) -> List[Metric]:
        """
        Load all metrics as plugins.
        By default, if no metrics is given, a default empty metric is loaded.

        :param metrics: List of metrics name to load.
        :return: Metrics classes loaded.
        """
        loaded_classes = []
        for metric in metrics:
            try:
                metric_module = importlib.import_module(f"metrics.{metric}")
                metric_class = None
                for obj in metric_module.__dict__.values():
                    if isinstance(obj, type) and Metric in obj.__bases__:
                        metric_class = obj
                        break
                if metric_class:
                    loaded_classes.append(metric_class())
                    logger.debug("Loaded metric class: %s", metric)
                else:
                    logger.error(
                        "No class inheriting from Metric found in metrics "
                        "module %s",
                        metric,
                    )
            except (ImportError, AttributeError) as e:
                logger.error("Error loading metric class %s: %s", metric, e)
        return loaded_classes

    def compute_dashboard(self, data: MetricData) -> List[dbc.Row]:
        """
        Compute widgets of the dashboard related to all metrics.

        :param data: Metric input data.
        :return: Dash widgets.
        """
        metrics_dashboard: list = []
        for metric in self._metrics:
            metrics_dashboard.extend(metric.compute_dashboard(data))
        return metrics_dashboard

    def compute_report(self, data: MetricData) -> List[MetricReport]:
        """
        Compute report related to all metrics.

        :param data: Metric input data.
        :return: Report of all metrics.
        """
        metrics_report: list = []
        for metric in self._metrics:
            metrics_report.append(metric.compute_report(data))
        return metrics_report



#! python3

"""
Unit tests for jiraclient.
"""

import pytest

from rate_my_project.connectors import JiraClient


def test_create_jira_client_with_empty_url() -> None:
    """
    Jira client creation with invalid url must raise an exception.
    """
    with pytest.raises(Exception):
        JiraClient("", "user", "pass")


def test_create_jira_client_with_empty_authentication() -> None:
    """
    Jira client creation with invalid url must raise an exception.
    """
    with pytest.raises(Exception):
        JiraClient("http://test", "", "pass")


def test_create_jira_client_with_empty_password() -> None:
    """
    Jira client creation with invalid password must raise an exception.
    """
    with pytest.raises(Exception):
        JiraClient("http://test", "user", "")



#! python3

"""
Manage command line interface.
"""

import logging
import sys

import click
from pydantic import ValidationError

from .config import load_global_config
from .dashboard import start_web_application
from .report import generate_all_reports


@click.group(context_settings={"help_option_names": ["-h", "--help"]})
@click.option("-v", "--verbose", is_flag=True, help="Enables verbose mode.")
def cli(verbose: bool = False) -> None:
    """
    Swiss knife for measuring project efficiency.
    """
    # Create logger
    if verbose:
        log_level = logging.DEBUG
    else:
        log_level = logging.INFO

    # Create logger
    logging.basicConfig(
        stream=sys.stdout,
        format="%(asctime)s - %(levelname)s - %(message)s",
        level=log_level,
    )


@cli.command()
@click.argument("config", type=click.Path(exists=True))
def report(config) -> None:
    """
    Generate report from CONFIG file.
    """
    # Get configuration from JSON/YAML file
    try:
        global_config = load_global_config(config)
    except ValidationError as error:
        sys.exit(str(error))

    logging.getLogger().debug(global_config.json())
    generate_all_reports(global_config)


@cli.command()
@click.argument("config", type=click.Path(exists=True))
def explore(config) -> None:
    """
    Explore efficiency metrics with web interface from CONFIG file.
    """
    # Get configuration from JSON/YAML file
    try:
        global_config = load_global_config(config)
    except ValidationError as error:
        sys.exit(str(error))

    logging.getLogger().debug(global_config.json())
    start_web_application(global_config)



#! python3

"""
Manage configuration file.
"""

from dataclasses import dataclass
import json
import logging
import pathlib
from typing import List, Optional

from pydantic import BaseModel, ConfigDict, Field, HttpUrl, SecretStr
from pydantic_settings import BaseSettings, SettingsConfigDict
import yaml

from .connectors import ConfluenceClient, JiraClient

#: Create logger for this file.
logger = logging.getLogger()


class Secrets(BaseSettings):
    """
    This class is used to store all secrets from environment variables.
    """

    #: Specific configuration
    model_config = SettingsConfigDict(
        frozen=False, env_file=".env", env_file_encoding="utf-8"
    )

    #: Username for Jira/Confluence
    user: str = Field(alias="ATLASSIAN_USER")
    #: Token for Jira/Confluence
    token: SecretStr = Field(alias="ATLASSIAN_TOKEN")


class ImmutableModel(BaseModel):
    """
    This class is used to have immutable model. It is used as base madel.
    """

    #: Specific configuration
    model_config = ConfigDict(frozen=False)


class Server(ImmutableModel):
    """
    This class is used to store server configuration.
    """

    #: Jira URL
    jira: HttpUrl

    #: Confluence URL
    confluence: HttpUrl


class Fields(ImmutableModel):
    """
    This class is used to store fields configuration.
    """

    #: Project name
    sprint: str
    #: Query to retrieve tickets
    story_points: str


class Report(ImmutableModel):
    """
    This class is used to store report configuration.
    """

    #: Confluence output space.
    space: str
    #: Confluence parent page.
    parent_page: str
    #: Page template to use.
    template: Optional[str] = Field("report.jinja2")


class WorkflowState(ImmutableModel):
    """
    This class is used to store report configuration.
    """

    #: Workflow state name.
    name: str
    #: Workflow state status to map.
    status: List[str]
    #: Start tag associated to this workflow state to compute lead time.
    start: Optional[bool] = Field(False)
    #: Stop tag associated to this workflow state to compute lead time.
    stop: Optional[bool] = Field(False)


class Project(ImmutableModel):
    """
    This class is used to store project configuration.
    """

    #: Project name
    name: str
    #: Query to retrieve tickets
    jql: str
    #: Report configuration.
    report: Report
    #: Workflow configuration.
    workflow: List[WorkflowState]

    def workflow_to_dict(self) -> list:
        """
        Returns the workflow configuration in list of dictionary.

        :return: Configuration string in JSON format.
        """
        return [workflow.model_dump() for workflow in self.workflow]


class Config(ImmutableModel):
    """
    This class is used to store main configuration excepted secrets.
    """

    #: Server configuration.
    server: Server
    #: List of metrics
    metrics: List[str]
    #: Fields configuration
    fields: Fields
    #: List of all projects
    projects: List[Project]


@dataclass
class GlobalConfig:
    """
    Global configuration.
    """

    #: Secrets configuration.
    secrets: Secrets
    # Main configuration.
    config: Config

    def json(self) -> str:
        """
        Returns the configuration in json format.

        :return: Configuration string in JSON format.
        """
        return f"""
        {
            {
                "secrets": {self.secrets.model_dump_json()},
                "config": {self.config.model_dump_json()}
            }
        }"""

    def confluence_client(self) -> ConfluenceClient:
        """
        Create and return a Jira client.

        :return: Jira client.
        """
        return ConfluenceClient(
            str(self.config.server.confluence),
            self.secrets.user,
            self.secrets.token.get_secret_value(),
        )

    def jira_client(self) -> JiraClient:
        """
        Create and return a Jira client.

        :return: Jira client.
        """
        return JiraClient(
            str(self.config.server.jira),
            self.secrets.user,
            self.secrets.token.get_secret_value(),
        )


def _parse_yaml_config(yaml_config_file: str) -> dict:
    """
    Constructs the configuration from YAML file.

    :param yaml_config_file: YAML configuration file to parse.
    :raises Exception: If configuration file is invalid.
    """
    logger.info("Parse YAML configuration from %s", yaml_config_file)

    try:
        with open(yaml_config_file, encoding="utf-8") as yaml_config:
            return yaml.safe_load(yaml_config)
    except yaml.YAMLError as error:
        raise ValueError("Failed to parse YAML configuration") from error


def _parse_json_config(json_config_file: str) -> dict:
    """
    Constructs the configuration from JSON file.

    :param json_config_file: JSON configuration file to parse.
    :raises Exception: If configuration file is invalid.
    """
    logger.info("Parse JSON configuration from %s", json_config_file)

    try:
        with open(json_config_file, encoding="utf-8") as json_config:
            return json.load(json_config)
    except json.JSONDecodeError as error:
        raise ValueError("Failed to parse JSON configuration") from error


def load_global_config(config_file: str) -> GlobalConfig:
    """
    Loads the configuration file (JSON or YAML) and the secrets.

    :param config_file: Configuration file to parse.
    :return: Configuration parsed.
    :raises Exception: If configuration extension file is unknown (.json,
    .yaml, .yml).
    :raises ValidationError: If configuration is invalid.
    """
    config_type = pathlib.Path(config_file).suffix
    if config_type in [".yaml", ".yml"]:
        config = _parse_yaml_config(config_file)
    elif config_type == ".json":
        config = _parse_json_config(config_file)
    else:
        raise ValueError("Unknown file extension for configuration")
    return GlobalConfig(Secrets(), Config.model_validate(config))



#! python3

"""
Client to communicate with Jira.
"""

import asyncio
import logging
from typing import List, Optional, Union

from aiohttp import BasicAuth

from .httpclient import HttpClient

#: Create logger for this file.
logger = logging.getLogger()


class JiraClient:
    """
    This class is used to interfacing Jira server.
    """

    #: Version of Atlassian API used
    API_VERSION: int = 3

    #: Standard headers
    STANDARD_HEADERS: dict = {
        "Content-Type": "application/json",
        "Accept": "application/json",
    }

    def __init__(
        self,
        jira_url: str,
        jira_username: str,
        jira_password: str,
    ):
        """
        Constructs the Jira client.

        :param jira_url: URL to connect to Jira.
        :param jira_username: Username to connect to Jira.
        :param jira_password: Password to connect to Jira.
        :raises Exception: If Jira credentials or URL are empty.
        """
        logger.debug("Create Jira client")

        if not jira_url:
            raise ValueError("Jira URL is invalid")
        if not jira_username:
            raise ValueError("Jira username is invalid")
        if not jira_password:
            raise ValueError("Jira password is invalid")

        self._http_client: HttpClient = HttpClient(
            jira_url.rstrip("/") + f"/rest/api/{self.API_VERSION}/",
            BasicAuth(jira_username, jira_password),
        )

        logger.debug("Jira client created")

    async def __aenter__(self):
        """
        Create session to send requests.
        """
        await self._http_client.__aenter__()
        return self

    async def __aexit__(self, *err):
        """
        Close session.
        """
        await self._http_client.__aexit__(*err)

    async def _get_paginated(
        self, suffix_url: str, headers: dict, query: dict, result_field: str
    ):
        """
        Send a GET request and aggregated the response.

        :param suffix_url: Last part of the URL contains the request.
        :param headers: Header for the request.
        :param query: Query of the request.
        :return: Response.
        """
        responses = []
        while True:
            response = await self._http_client.get(suffix_url, headers, query)
            responses += response[result_field]

            if query["startAt"] + query["maxResults"] >= response["total"]:
                break
            query["startAt"] += query["maxResults"]

        return responses

    async def ticket(
        self, key: str, fields: Optional[Union[List[str], str]] = None
    ) -> dict:
        """
        Get ticket information from a given `key`.

        :param key: Key of the ticket.
        :param fields: list of fields, for example: ['priority', 'summary']
        :return: Ticket information.
        """
        if fields is None:
            fields = "*all"

        query = {
            "fields": fields,
        }
        return await self._http_client.get(
            f"issue/{key}", self.STANDARD_HEADERS, query
        )

    async def validate_jql(self, jql: str):
        """
        Validate `jql` request.

        :param jql: JQL request to validate.
        """
        query = {"queries": [jql]}
        response = await self._http_client.post(
            "jql/parse", self.STANDARD_HEADERS, query
        )
        if "errors" in response["queries"][0]:
            raise ValueError(response["queries"][0]["errors"][0])

    async def tickets_from_jql(
        self, jql: str, fields: Optional[Union[List[str], str]] = None
    ) -> list:
        """
        Get tickets from a `jql` request.

        :param jql: JQL request to find tickets.
        :param fields: list of fields, for example: ['priority', 'summary']
        :return: Tickets list found.
        """
        if fields is None:
            fields = "*all"

        query = {
            "jql": jql,
            "fields": fields,
            "startAt": 0,
            "maxResults": 100,
        }
        return await self._get_paginated(
            "search", self.STANDARD_HEADERS, query, "issues"
        )

    async def changelogs(self, key: str) -> list:
        """
        Get all changelogs of a given ticket.

        :param key: Key of the ticket.
        :return: Changelogs list.
        """
        query = {
            "startAt": 0,
            "maxResults": 100,
        }
        return await self._get_paginated(
            f"issue/{key}/changelog", self.STANDARD_HEADERS, query, "values"
        )

    async def changelogs_from_tickets(self, tickets: list):
        """
        Get changelogs information from a list of `tickets`.

        :param tickets: List of tickets.
        :return: List of changelogs information.
        """
        tasks = []
        for ticket in tickets:
            task = asyncio.create_task(self.changelogs(ticket["key"]))
            tasks.append(task)

        return await asyncio.gather(*tasks)

    async def parents_from_tickets(self, tickets: list) -> list:
        """
        Get parents information from a list of `tickets`.

        :param tickets: List of tickets.
        :return: List of parents information.
        """
        tasks = []
        for ticket in tickets:
            if ticket["fields"].get("parent"):
                task = asyncio.create_task(
                    self.ticket(ticket["fields"]["parent"]["key"])
                )
                tasks.append(task)

        return await asyncio.gather(*tasks)

    async def versions(self, key: str) -> list:
        """
        Get all versions of a given project ordered by ranking.

        :param key: Key of the project.
        :return: Versions list.
        """
        query = {"startAt": 0, "maxResults": 100, "orderBy": "-sequence"}
        return await self._get_paginated(
            f"project/{key}/version", self.STANDARD_HEADERS, query, "values"
        )

    async def fields_information(self):
        """
        Get all fields information like id and associated display name.

        :return: List of fields information.
        """
        return await self._http_client.get("field", self.STANDARD_HEADERS)



#! python3

"""
Unit tests for config.
"""

import os

import pytest

from rate_my_project.config import GlobalConfig, load_global_config


@pytest.fixture(autouse=True)
def secrets():
    """
    Set secrets as environment variables.
    """

    os.environ["ATLASSIAN_USER"] = "Username"
    os.environ["ATLASSIAN_TOKEN"] = "Token"


@pytest.fixture(scope="module")
def script_loc(request):
    """
    Return the directory of the currently running test script.
    """

    # uses .join instead of .dirname so we get a LocalPath object instead of
    # a string. LocalPath.join calls normpath for us when joining the path
    return request.fspath.join("..")


def test_yaml_basic_config(script_loc) -> None:
    """
    YAML basic config must be loaded without errors.
    """
    config = script_loc.join("../examples/basic_config.yaml")
    load_global_config(config)


def test_json_basic_config(script_loc) -> None:
    """
    JSON basic config must be loaded without errors.
    """
    config = script_loc.join("../examples/basic_config.json")
    load_global_config(config)


def test_yaml_multi_projects_config(script_loc) -> None:
    """
    YAML multi projects config must be loaded without errors.
    """
    config = script_loc.join("../examples/multi_projects_config.yaml")
    load_global_config(config)


def test_json_multi_projects_config(script_loc) -> None:
    """
    JSON multi projects config must be loaded without errors.
    """
    config = script_loc.join("../examples/multi_projects_config.json")
    load_global_config(config)


def test_yaml_multi_projects_with_anchor_config(script_loc) -> None:
    """
    YAML multi projects config with_anchor must be loaded without errors.
    """
    config = script_loc.join(
        "../examples/multi_projects_with_anchor_config.yaml"
    )
    load_global_config(config)


def test_config_with_empty_config() -> None:
    """
    An empty config must raise an exception.
    """
    with pytest.raises(Exception):
        GlobalConfig({})



#! python3

"""
Helper functions.
"""

import logging

import inflection

from .connectors import JiraClient

#: Create logger for this file.
logger = logging.getLogger()


async def fetch_tickets_information(
    jira_client: JiraClient, jql: str
) -> tuple:
    """
    Fetch tickets and associated changelogs information from `JQL` query.

    TODO The fields list will be fetched to retrieve custom fields if needed.

    :param jira_client: Jira client to retrieve tickets information.
    :param jql: JQL query.
    :return: Tickets, changelogs and parents tickets.
    """
    async with jira_client as jira_session:
        tickets = await jira_session.tickets_from_jql(jql)
        changelogs = await jira_session.changelogs_from_tickets(tickets)
        parents = await jira_session.parents_from_tickets(tickets)
        # fields = await jira_session.fields_information()
        return tickets, changelogs, parents


def to_snake_case(text: str) -> str:
    """
    Convert test to snake case.

    :param text: Input text to convert.
    :return: Text in snake case.
    """
    # underscore function do not replace the space
    return inflection.underscore(text.replace(" ", "_"))



#! python3

"""
Confluence report generation.
"""

import asyncio
import logging
import os

from jinja2 import Environment, ChoiceLoader, FileSystemLoader, PackageLoader

from .config import GlobalConfig
from .connectors import ConfluenceClient
from .metric import Metrics, MetricData
from .utils import fetch_tickets_information

#: Create logger for this file.
logger = logging.getLogger()

#: Object to manipulate the templates.
_jinja_environment: Environment = Environment(
    loader=ChoiceLoader(
        [
            FileSystemLoader(searchpath=os.getcwd()),
            PackageLoader("rate_my_project.metrics"),
        ]
    ),
    autoescape=True,
)


def generate_report(
    confluence_client: ConfluenceClient,
    project_name: str,
    output_space: str,
    output_parent_page: str,
    template: str,
    metric_reports: list,
) -> None:
    """
    Generate the report for the `project_name`.
    The report will be generated on the Confluence `output_space` as a child of
    `output_parent_page`. The page will be generated by taking the given
    `template`.
    The `confluence_client` is used to get the template and publish the report.

    :param confluence_client: Confluence client to publish report.
    :param project_name: Project name.
    :param output_space: Project output Confluence space.
    :param output_parent_page: Project output Confluence parent page.
    :param template: Template page used to render the report.
    :param metric_reports: Data for each metrics needed to generate report.
    """
    logger.debug("Generate report for %s", project_name)

    # Generate report from template
    jinja_template = _jinja_environment.get_template(template)
    content = jinja_template.render(metric_reports=metric_reports)
    report = content.encode("ascii", "ignore").decode()

    # Publish report
    asyncio.run(
        confluence_client.create_or_update_page(
            space=output_space,
            parent_page=output_parent_page,
            title=project_name,
            message=report,
        )
    )

    # Add figures to the report
    for metric_report in metric_reports:
        if metric_report.figures:
            asyncio.run(
                confluence_client.upload_files(
                    space=output_space,
                    title=project_name,
                    filenames=metric_report.figures,
                )
            )

    logger.debug("Report for %s generated", project_name)


def generate_all_reports(global_config: GlobalConfig) -> None:
    """
    Generate all reports given by `global_config`.

    :param global_config: Global configuration.
    """
    logger.info("Generate report on Confluence")

    for project in global_config.config.projects:
        jira_information = asyncio.run(
            fetch_tickets_information(global_config.jira_client(), project.jql)
        )
        metric_data = MetricData.from_tickets_and_changelogs(
            global_config.config.fields.model_dump(),
            project.workflow_to_dict(),
            *jira_information,
        )

        metrics = Metrics(global_config.config.metrics)
        reports = metrics.compute_report(metric_data)

        generate_report(
            global_config.confluence_client(),
            project.name,
            project.report.space,
            project.report.parent_page,
            project.report.template,
            reports,
        )

    logger.info("Report on Confluence generated")



#! python3

"""
Unit tests for utils.
"""

from rate_my_project.utils import to_snake_case


def test_one_word_already_in_snake_case_to_snake_case() -> None:
    """
    Convert one word without spaces in snake case to snake case.
    """
    assert to_snake_case("word_in_snake_case") == "word_in_snake_case"


def test_one_word_to_snake_case() -> None:
    """
    Convert one word without spaces not in snake case to snake case.
    """
    assert to_snake_case("WordInCamelCase") == "word_in_camel_case"


def test_multiple_words_with_spaces_to_snake_case() -> None:
    """
    Convert multiple words separated by spaces not in snake case to snake case.
    """
    assert to_snake_case("Word with spaces") == "word_with_spaces"



#! python3

"""
Client to communicate with Confluence.
"""

import logging
from typing import List

from aiohttp import BasicAuth

from .httpclient import HttpClient

#: Create logger for this file.
logger = logging.getLogger()


class ConfluenceClient:
    """
    This class is used to interfacing Confluence server.
    """

    #: Standard headers
    STANDARD_HEADERS: dict = {
        "Content-Type": "application/json",
        "Accept": "application/json",
    }

    #: No check headers
    NO_CHECK_HEADERS: dict = {
        "X-Atlassian-Token": "no-check",
    }

    def __init__(
        self,
        confluence_url: str,
        confluence_username: str,
        confluence_password: str,
    ):
        """
        Constructs the Confluence client.

        :param confluence_url: URL to connect to Confluence.
        :param confluence_username: Username to connect to Confluence.
        :param confluence_password: Password to connect to Confluence.
        :raises Exception: If URL, username or password are invalids.
        """
        logger.debug("Create Confluence client")

        if not confluence_url:
            raise ValueError("Confluence URL is invalid")
        if not confluence_username:
            raise ValueError("Confluence username is invalid")
        if not confluence_password:
            raise ValueError("Confluence password is invalid")

        self._http_client: HttpClient = HttpClient(
            confluence_url.rstrip("/") + "/wiki/rest/api/",
            BasicAuth(confluence_username, confluence_password),
        )

        logger.debug("Confluence client created")

    async def __aenter__(self):
        """
        Create session to send requests.
        """
        await self._http_client.__aenter__()
        return self

    async def __aexit__(self, *err):
        """
        Close session.
        """
        await self._http_client.__aexit__(*err)

    async def get_page_content(self, space: str, title: str) -> dict:
        """
        Get a page content from `title` in the `space`.

        :param space: Confluence space.
        :param title: Page title.
        :return: Content if page is found.
        """
        query = {
            "spaceKey": space,
            "title": title,
        }
        response = await self._http_client.get(
            "content", self.STANDARD_HEADERS, query
        )
        if not response["results"]:
            raise ValueError("Page not found")
        return response["results"][0]

    async def get_page_content_id(self, space: str, title: str) -> int:
        """
        Get a page content identifier from `title` in the `space`.

        :param space: Confluence space.
        :param title: Page title.
        :return: Identifier of the page.
        """
        response = await self.get_page_content(space, title)
        return response["id"]

    async def get_page_content_version(self, space: str, title: str) -> int:
        """
        Get a page content version from `title` in the `space`.

        :param space: Confluence space.
        :param title: Page title.
        :return: Version of the page.
        """
        page_id = await self.get_page_content_id(space, title)
        response = await self._http_client.get(
            f"content/{page_id}/history", self.STANDARD_HEADERS
        )
        return response["lastUpdated"]["number"]

    async def create_or_update_page(
        self,
        space: str,
        parent_page: str,
        title: str,
        message: str,
        representation="wiki",
    ) -> None:
        """
        Create or update page in Confluence in the given `space`. The page will
        be located under the `parent_page` and will have the given `title` and
        content `message` in given `representation` format. By default,
        Wiki markup format is used but "storage" can also be used.

        :param space: Confluence space.
        :param parent_page: Name of the parent page.
        :param title: Page title.
        :param message: Page message.
        :param representation: Content format (wiki or storage).
        :raises Exception: If parent page is not found.
        """
        logger.debug("Create or update page")

        if representation not in ["storage", "wiki"]:
            raise ValueError("Wrong value for representation")

        parent_id = await self.get_page_content_id(space, parent_page)

        query = {
            "type": "page",
            "title": title,
            "space": {"key": space},
            "ancestors": [{"id": parent_id}],
            "body": {
                representation: {
                    "value": message,
                    "representation": representation,
                }
            },
        }

        update = False
        try:
            page_id = await self.get_page_content_id(space, title)
            query["id"] = str(page_id)
            update = True
        except Exception:
            pass

        if update:
            version = await self.get_page_content_version(space, title)
            query["version"] = {"number": version + 1}
            await self._http_client.put(
                f"content/{query['id']}",
                self.STANDARD_HEADERS,
                query,
            )
            logger.debug("Page updated")
        else:
            await self._http_client.post(
                "content", self.STANDARD_HEADERS, query
            )
            logger.debug("New page created")

    async def upload_files(
        self,
        space: str,
        title: str,
        filenames: List[str],
    ):
        """
        Upload files and attach to the page with given `title` in the `space`.

        :param space: Confluence space.
        :param title: Page title.
        :param filenames: List of file to upload.
        """
        page_id = await self.get_page_content_id(space, title)
        for filename in filenames:
            with open(filename, "rb") as file:
                query = {
                    "type": "attachment",
                    "minorEdit": "true",
                    "file": file,
                }

                await self._http_client.put(
                    f"content/{page_id}/child/attachment",
                    self.NO_CHECK_HEADERS,
                    data=query,
                )



#! python3

"""
Web dashboard to explore metrics.
"""

import asyncio

from dash import dcc, Dash, DiskcacheManager, html, Input, Output, State
from dash.exceptions import PreventUpdate
import dash_bootstrap_components as dbc
from dash_bootstrap_templates import load_figure_template
import diskcache

# from werkzeug.middleware.profiler import ProfilerMiddleware

from .config import GlobalConfig
from .metric import Metrics, MetricData
from .utils import fetch_tickets_information

# local cache to store query results
cache = diskcache.FanoutCache()
# Jira client to do the requests
CONFIG: GlobalConfig


def create_application_layout() -> html.Div:
    """
    Create layout of dashboard application dashboard.
    """
    navbar = dbc.NavbarSimple(
        children=[],
        brand="Rate My Project",
        brand_href="#",
        color="primary",
        dark=True,
        className="mb-3",
    )

    search = dbc.Container(
        [
            dbc.Row(
                [
                    dbc.Col(
                        [
                            dbc.Input(
                                id="jql_input",
                                type="text",
                                placeholder="Enter your JQL query here",
                            ),
                            dbc.FormFeedback(
                                "Valid JQL query",
                                id="jql_ok",
                                type="valid",
                            ),
                            dbc.FormFeedback(
                                "", id="jql_error", type="invalid"
                            ),
                        ]
                    ),
                    dbc.Col(
                        dbc.Button("Run Analysis", id="run_button"),
                        width="auto",
                    ),
                ],
            ),
        ],
        className="mb-3",
    )

    return html.Div(
        [
            # List of tickets retrieved with last JQL query.
            dcc.Store(id="tickets", storage_type="session"),
            navbar,
            search,
            dbc.Container(id="metrics"),
        ],
    )


def create_application() -> Dash:
    """
    Create dashboard application and define the layout of the dashboard.
    """
    app = Dash(
        external_stylesheets=[dbc.themes.COSMO],
        background_callback_manager=DiskcacheManager(cache, expire=60),
    )
    load_figure_template("cosmo")
    app.layout = create_application_layout()
    return app


application = create_application()


@cache.memoize(typed=True, expire=60)
def fetch_data_for_metric_from_jql(jql: str) -> MetricData:
    """
    Fetch the tickets from `JQL` query and create data used by metric.
    The data is cached during 60s to avoid too many requests to the server.

    :param jql: JQL query.
    :return: Data for metric.
    """
    jira_information = asyncio.run(
        fetch_tickets_information(CONFIG.jira_client(), jql)
    )
    return MetricData.from_tickets_and_changelogs(
        CONFIG.config.fields.dict(),
        CONFIG.config.projects[0].workflow_to_dict(),
        *jira_information,
    )


@application.callback(
    inputs=Input("jql_input", "n_submit"),
    state=State("run_button", "n_clicks"),
    output=Output("run_button", "n_clicks"),
)
def on_jql_input_entered(jql, run_button_clicks):
    """
    When the JQL is validated with enter, it runs the analysis by simulating a
    click on the run button.
    """
    if not jql:
        raise PreventUpdate

    if run_button_clicks:
        return run_button_clicks + 1
    return 1


@application.callback(
    inputs=Input("run_button", "n_clicks"),
    state=State("jql_input", "value"),
    output=Output("metrics", "children"),
    background=True,
    running=[
        (Output("run_button", "disabled"), True, False),
    ],
)
def on_run_button_clicked(_, jql):
    """
    Execute the analysis when the run button is clicked by fetching the tickets
    from JQL query.
    """
    if not jql:
        raise PreventUpdate
    metrics = Metrics(CONFIG.config.metrics)
    tickets = fetch_data_for_metric_from_jql(jql)
    return metrics.compute_dashboard(tickets)


@application.callback(
    inputs=Input("jql_input", "value"),
    output=[
        Output("jql_input", "valid"),
        Output("jql_input", "invalid"),
        Output("jql_error", "children"),
    ],
)
def on_jql_input_updated(jql):
    """
    Validate the JQL query syntax.
    """
    if not jql:
        return False, False, ""
    try:
        asyncio.run(CONFIG.jira_client().validate_jql(jql))
        return True, False, ""
    except Exception as error:
        return False, True, str(error)


def start_web_application(global_config: GlobalConfig) -> None:
    """
    Start dashboard server.
    """
    global CONFIG
    CONFIG = global_config
    # Profiling
    # application.server.config["PROFILE"] = True
    # application.server.wsgi_app = ProfilerMiddleware(
    #     application.server.wsgi_app, sort_by=("cumtime", "tottime"),
    #     restrictions=[50]
    # )
    application.run_server(debug=False)



#! python3

"""
Unit tests for httpclient.
"""

import pytest
from aiohttp import BasicAuth

from rate_my_project.connectors import HttpClient


def test_create_http_client_with_empty_url() -> None:
    """
    Http client creation with invalid url must raise an exception.
    """
    with pytest.raises(Exception):
        HttpClient("", BasicAuth("user", "pass"))


def test_create_jira_client_with_empty_username() -> None:
    """
    Http client creation with no authentication must not raise an exception.
    """
    HttpClient("http://test", None)



#! python3

"""
Status chart.
"""
import logging
from typing import List

import dash_bootstrap_components as dbc
import plotly.express as px

from rate_my_project.metric import Metric, MetricData, MetricReport
from rate_my_project.utils import to_snake_case

#: Create logger for this file.
logger = logging.getLogger()


class Status(Metric):
    """
    This class is used to provide metric about status.
    """

    def compute_dashboard(self, data: MetricData) -> List[dbc.Row]:
        """
        Interface to compute widget of the dashboard related to this metric.

        :param data: Metric input data.
        :return: Dash widgets.
        """
        figure_list = []
        for issue_type, figure in self.status_per_issue_type(data).items():
            figure_list.append(
                self._draw_figure(f"Status per {issue_type}", figure)
            )
        return [dbc.Row(figure_list, class_name="row g-2")]

    def compute_report(self, data: MetricData) -> MetricReport:
        """
        Interface to compute widget of the report related to this metric.

        :param data: Metric input data.
        :return: Report of the metric.
        """
        figures = []
        issue_types = []
        for issue_type, figure in self.status_per_issue_type(data).items():
            issue_type_snake_case = to_snake_case(issue_type)
            figure_path = f"{self.OUTPUT_DIR}/{issue_type_snake_case}.png"
            figure.write_image(figure_path)
            figures.append(figure_path)
            issue_types.append(
                {"type": issue_type, "path": issue_type_snake_case}
            )
        return MetricReport(
            metric_name="Status",
            figures=figures,
            report_data={"issue_types": issue_types},
        )

    @staticmethod
    def status_per_issue_type(data: MetricData) -> dict:
        """
        Compute the number of tickets in a status for each issue type.

        :param data: Metric input data.
        :return: Dictionary with a graph for each issue type name.
        """
        result = {}

        groups = data.changes.merge(data.info, on="key").groupby("type")

        for issue_type, group in groups:
            df_per_type = group[["key", "status"]].drop_duplicates(
                subset="key", keep="last"
            )
            status_per_type = (
                df_per_type["status"]
                .cat.remove_unused_categories()
                .value_counts(sort=False)
            )

            result[issue_type] = px.pie(
                values=status_per_type, names=status_per_type.index, hole=0.4
            )

        return result



#! python3

"""
Unit tests for confluenceclient.
"""

import pytest

from rate_my_project.connectors import ConfluenceClient


def test_create_confluence_client_with_empty_url() -> None:
    """
    Confluence client creation with invalid url must raise an exception.
    """
    with pytest.raises(Exception):
        ConfluenceClient("", "user", "pass")


def test_create_confluence_client_with_empty_username() -> None:
    """
    Confluence client creation with invalid username must raise an exception.
    """
    with pytest.raises(Exception):
        ConfluenceClient("http://test", "", "pass")


def test_create_confluence_client_with_empty_password() -> None:
    """
    Confluence client creation with invalid password must raise an exception.
    """
    with pytest.raises(Exception):
        ConfluenceClient("http://test", "user", "")



#! python3

"""
Module of all connectors used to collect or write data.
"""

from .confluenceclient import ConfluenceClient
from .httpclient import HttpClient
from .jiraclient import JiraClient

__all__ = ["ConfluenceClient", "HttpClient", "JiraClient"]



#! python3

"""
Client to communicate with Http.
"""

import logging
from typing import Any, Optional

from aiohttp import BasicAuth, ClientSession, TCPConnector

#: Create logger for this file.
logger = logging.getLogger()


class HttpClient:
    """
    This class is used to interfacing Http server.
    """

    def __init__(
        self,
        http_url: str,
        auth: Optional[BasicAuth] = None,
    ):
        """
        Constructs the Http client.

        :param http_url: URL to connect to Http server.
        :param auth: Authentication to connect to server.
        :raises Exception: If URL is empty.
        """
        logger.debug("Create Http client")

        if not http_url:
            raise ValueError("Http URL is invalid")
        self._url: str = http_url

        #: Server authentication
        self._auth: Optional[BasicAuth] = auth

        #: HTTP session
        self._session: Optional[ClientSession] = None

        logger.debug("Http client created")

    async def __aenter__(self):
        """
        Create Http session to send multiple requests.
        """
        self._session = ClientSession(
            auth=self._auth,
            raise_for_status=True,
            connector=TCPConnector(limit=50),
        )
        return self

    async def __aexit__(self, *err):
        """
        Close Http session.
        """
        await self._session.close()
        self._session = None

    async def get(self, suffix_url: str, headers: dict, query: Any = None):
        """
        Send a GET request.

        :param suffix_url: Last part of the URL contains the request.
        :param headers: Header for the request.
        :param query: Query of the request.
        :return: Response.
        """
        if self._session:
            async with self._session.get(
                url=self._url + suffix_url,
                headers=headers,
                params=query,
            ) as response:
                return await response.json()
        else:
            async with ClientSession(
                auth=self._auth,
                raise_for_status=True,
                connector=TCPConnector(limit=50),
            ) as session:
                async with session.get(
                    url=self._url + suffix_url,
                    headers=headers,
                    params=query,
                ) as response:
                    return await response.json()

    async def post(
        self,
        suffix_url: str,
        headers: dict,
        json: Any = None,
        data: Any = None,
    ):
        """
        Send a POST request.

        :param suffix_url: Last part of the URL contains the request.
        :param headers: Header for the request.
        :param json: JSON payload of the request.
        :param data: Payload of the request.
        :return: Response.
        """
        if self._session:
            async with self._session.post(
                url=self._url + suffix_url,
                data=data,
                json=json,
                headers=headers,
            ) as response:
                return await response.json()
        else:
            async with ClientSession(
                auth=self._auth,
                raise_for_status=True,
                connector=TCPConnector(limit=50),
            ) as session:
                async with session.post(
                    url=self._url + suffix_url,
                    data=data,
                    json=json,
                    headers=headers,
                ) as response:
                    return await response.json()

    async def put(
        self,
        suffix_url: str,
        headers: dict,
        json: Any = None,
        data: Any = None,
    ):
        """
        Send a PUT request.

        :param suffix_url: Last part of the URL contains the request.
        :param headers: Header for the request.
        :param json: JSON payload of the request.
        :param data: Payload of the request.
        :return: Response.
        """
        if self._session:
            async with self._session.put(
                url=self._url + suffix_url,
                data=data,
                json=json,
                headers=headers,
            ) as response:
                return await response.json()
        else:
            async with ClientSession(
                auth=self._auth,
                raise_for_status=True,
                connector=TCPConnector(limit=50),
            ) as session:
                async with session.put(
                    url=self._url + suffix_url,
                    data=data,
                    json=json,
                    headers=headers,
                ) as response:
                    return await response.json()
